/* chilunfold.c -- unfold NetCDF velocity file */

#include <stdlib.h>
#include <stdio.h>
#include <netcdf.h>
#include <string.h>

#include "dan.h"
#include "nctools.h"

#define NO_SIGNAL -999.00
#define NO_FLOAT_ATTRIBUTE -1.0e30
#define NO_FLOAT_ATTRIBUTE_TEST -0.9e30

static int ncstatus;
#define handle_error(a) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s\n", __LINE__, \
      nc_strerror(ncstatus)); nc_close(ncid); exit(-1); \
}
#define report_error(a, s) if ((ncstatus = (a)) != NC_NOERR) { \
    fprintf(stderr, "NetCDF error on line %d: %s (%s)\n", \
	    __LINE__, nc_strerror(ncstatus), s); exit(-1); \
}

static
char *
get_char_attribute(int ncid, int varid, char *name)
{
  size_t length;
  nc_type type;
  char *value;

  report_error(nc_inq_atttype(ncid, varid, name, &type), name);
  if (type != NC_CHAR) {
    fprintf(stderr, "Error: attribute %s is not of type char\n", name);
    exit(-1);
  }
  report_error(nc_inq_attlen(ncid, varid, name, &length), name);
  value = malloc((length+1) * sizeof(char));
  if (!value) {
    fprintf(stderr, "Out of memory\n");
    exit(2);
  }
  report_error(nc_get_att_text(ncid, varid, name, value), name);
  value[length] = '\0';
  return value;
}

static
float
get_float_attribute(int ncid, int varid, char *name)
{
  size_t length;
  float value;

  if (nc_inq_attlen(ncid, varid, name, &length) != NC_NOERR) {
    return NO_FLOAT_ATTRIBUTE;
  }
  if (length > 1) {
    fprintf(stderr, "Error: %s attribute has more than 1 element\n", name);
  }
  handle_error(nc_get_att_float(ncid, varid, name, &value));
  return value;
}

static
void
Usage(char **argv)
{
  fprintf(stderr, "Usage: %s [options] file1.nc [file2.nc ...]\n", argv[0]);
  fprintf(stderr, "Option:\n"
	  "         -user s  name and email address of the user\n"
	  "   -noisyPHIDP    do not clean phi_dp\n");

}


int
main(int argc, char **argv)
{
  float nan = NO_SIGNAL;
  int iarg, ifile;
  float maxvar = 5.0, lonely1 = 4.0, lonely2 = 4.0;
  char *user = NULL;
  char noisyphidp = 0;

  /* Parse switches */
  iarg=1;
  while (iarg < argc && *argv[iarg] == '-') {
    if (strcmp(argv[iarg],"-user") == 0) {
      if (argc > iarg+1) {
	if (user) {
	  fprintf(stderr, "Warning: user specified more than once:"
		  " `%s' not used\n", user);
	}
	user = argv[++iarg];
      }
      else {
	fprintf(stderr, "Error: user not specified\n");
	exit(1);
      }
    }
    else if (strcmp(argv[iarg],"-noisyPHIDP") == 0) {
      noisyphidp = 1;
    }
    ++iarg;
  }
  if (iarg > argc-1) {
    Usage(argv);
    exit(0);
  }

  for (ifile = iarg; ifile < argc; ++ifile) {
    size_t ngates, nrays, len, len_unfolded;
    float *v_new = NULL, *v_old = NULL;
    float *phidp = NULL;
    int *region = NULL;
    char *long_name = NULL;
    char *scantype = NULL;
    char unfolded[] = "Unfolded ";
    float vfold, vfold2, vfold_dummy=100;
    int ncid, timedimid, rangedimid;
    int phidpid, v1id, v2id;
    int elevid;
    int isppi = 0, isreversed = 0;
    char have_phidp = 0, have_v1 = 0;
    char *ch;

    /* Open file and get dimension sizes */
    handle_error(nc_open(argv[ifile], NC_WRITE | NC_NOCLOBBER, &ncid));
    handle_error(nc_inq_dimid(ncid, "time", &timedimid));
    handle_error(nc_inq_dimlen(ncid, timedimid, &nrays));
    handle_error(nc_inq_dimid(ncid, "range", &rangedimid));
    handle_error(nc_inq_dimlen(ncid, rangedimid, &ngates));

    /* Find out what variables we have */
    if (nc_inq_varid(ncid, "v_unfolded", &v1id) == NC_NOERR) {
      fprintf(stderr, "Skipping %s: already unfolded\n", argv[ifile]);
      continue;
    }
    if (nc_inq_varid(ncid, "v2", &v2id) != NC_NOERR) {
      fprintf(stderr, "Skipping %s: does not contain v2\n", argv[ifile]);
      continue;
    }
    if (nc_inq_varid(ncid, "v1", &v1id) == NC_NOERR) {
      have_v1 = 1;
    }
    if (nc_inq_varid(ncid, "phi_dp", &phidpid) == NC_NOERR) {
      have_phidp = 1;
    }
    
    scantype = get_char_attribute(ncid, NC_GLOBAL, "scantype");
    if (strcmp(scantype, "PPI") == 0 || strcmp(scantype, "Fixed") == 0) {
      isppi = 1;
    }
    else if (nc_inq_varid(ncid, "elev", &elevid) == NC_NOERR) {
      size_t index;
      float elev1, elev2;
      index = 0;
      handle_error(nc_get_var1_float(ncid, elevid, &index, &elev1));
      index = nrays-1;
      handle_error(nc_get_var1_float(ncid, elevid, &index, &elev2));
      if (elev1 > elev2) {
	isreversed = 1;
      }
    }
    free(scantype);
    scantype = NULL;

    vfold = get_float_attribute(ncid, v2id, "folding_velocity");
    if (vfold < NO_FLOAT_ATTRIBUTE_TEST) {
      fprintf(stderr, "Skipping %s: folding velocity unknown\n",
	      argv[ifile]);
      nc_close(ncid);
      continue;
    }
    vfold2 = vfold*2;
    printf("Unfolding %s: ", argv[ifile]);

    /* Create arrays to hold data */
    v_old = (float *) malloc(sizeof(float) * nrays * ngates);
    v_new = (float *) malloc(sizeof(float) * nrays * ngates);
    region = (int *) malloc(sizeof(int) * nrays * ngates);

    if (have_phidp) {
      phidp = (float *) malloc(sizeof(float) * nrays * ngates);
    }
    
    if (!v_old || !v_new || !region
	|| (have_phidp && !phidp)) {
      fprintf(stderr, "Out of memory\n");
      nc_close(ncid);
      exit(1);
    }

    handle_error(nc_get_var_float(ncid, v2id, v_old));
    if (have_phidp) {
      handle_error(nc_get_var_float(ncid, phidpid, phidp));
    }

    if (isppi) {
      printf("PPI ");
    }
    else {
      printf("RHI ");
    }
    if (isreversed) {
      printf("(reversed) ");
    }

    munfill(v_new, region, v_old, &ngates, &nrays,
	    &vfold, &nan, &isppi, &isreversed);
    printf(" rezapping...");
    mzap(v_old, v_new, &maxvar, &lonely1, &lonely2,
	 &ngates, &nrays, &vfold_dummy, &nan);
    printf("\n");

    if (have_phidp && !noisyphidp) {
      long int i;
      /* Remove noise from PHIDP using the zapped v2 field */
      for (i = 0; i < nrays*ngates; i++) {
	if (v_old[i] == NO_SIGNAL) {
	  phidp[i] = NO_SIGNAL;
	}
      }
    }

    /* v1 will be where we put the new cleaned & unfolded velocity paramter */
    handle_error(nc_redef(ncid));
    if (!have_v1) {
      int dimids[2];
      dimids[0] = timedimid;
      dimids[1] = rangedimid;
      handle_error(nc_def_var(ncid, "v_unfolded", NC_FLOAT, 2, dimids, &v1id));
    }
    else {
      nc_rename_var(ncid, v1id, "v_unfolded");
      nc_del_att(ncid, v1id, "folding_velocity");
    }

    handle_error(nc_inq_attlen(ncid, v2id, "long_name", &len));
    len_unfolded = strlen(unfolded);
    long_name = malloc(len_unfolded + len + 1);
    if (!long_name) {
      fprintf(stderr, "Out of memory\n");
      nc_close(ncid);
      exit(1);      
    }
    sprintf(long_name, "%s", unfolded);
    handle_error(nc_get_att_text(ncid, v2id, "long_name",
				 long_name + len_unfolded));
    long_name[len_unfolded + len] = '\0';
    /* Remove any `(H&V)' or suchlike in the name */
    ch = long_name + len_unfolded + len - 1;
    if (*ch == ')') {
      while ((--ch) > long_name+len_unfolded) {
	if (*ch == '(') {
	  ch--;
	  if (*ch == ' ') {
	    *ch = '\0';
	  }
	  break;
	}
      }
    }

    handle_error(nc_put_att_text(ncid, v1id, "long_name",
				 strlen(long_name), long_name));
    handle_error(nc_put_att_text(ncid, v2id, "long_name",
				 strlen(long_name+len_unfolded),
				 long_name+len_unfolded));

    free(long_name);
    long_name = NULL;

    /* Rename v2 to v */
    nc_rename_var(ncid, v2id, "v");

    /* Copy comments from v2 to v1 */
    nc_copy_att(ncid, v2id, COMMENT_NAME, ncid, v1id);  

    /* Record what we've done */
    handle_error(nct_add_command_line(ncid, argc, argv));
    handle_error(nct_add_history(ncid, "Unfolded", user));
    handle_error(nct_append_attribute(ncid, v1id, COMMENT_NAME,
	      "Noise has been removed on the basis of the variance "
	      "of the velocity at each pixel with its neighbours.\n"
	      "This can occasionally remove good data with "
	      "genuinely high variance.\n" 
	      "Each contiguous region of data has been unfolded "
	      "using a flood-fill approach, and separated "
	      "regions made consistent based on their proximity.\n"
	      "Note that The unfolding algorithm sometimes unfolds "
	      "regions incorrectly.", "\n"));
    if (noisyphidp) {
      handle_error(nct_append_attribute(ncid, phidpid, COMMENT_NAME,
	      "Noise can be removed from this parameter by rejecting "
	      "pixels for which v_unfolded is equal to v_unfolded:missing_value.", "\n"));
    }
    else {
      handle_error(nct_append_attribute(ncid, phidpid, COMMENT_NAME,
	      "Noise has been removed on the basis of the variance "
	      "of the velocity at each pixel with its neighbours.\n"
	      "This can occasionally remove good data with "
	      "genuinely high variance.", "\n"));
    }
    handle_error(nct_append_attribute(ncid, v2id, COMMENT_NAME,
	      "This is the raw velocity measured by the radar, "
	      "which folds at +/- the value in the folding_velocity attribute.\n"
	      "Noise can be removed from this parameter by rejecting "
	      "pixels for which v_unfolded is equal to v_unfolded:missing_value.", "\n"));

    handle_error(nc_enddef(ncid));

    /* Insert variables */
    handle_error(nc_put_var_float(ncid, v1id, v_old));
    if (have_phidp) {
      handle_error(nc_put_var_float(ncid, phidpid, phidp));
    }

    handle_error(nc_put_att_float(ncid, v1id, "missing_value",
				  NC_FLOAT, 1, &nan));
    handle_error(nc_put_att_float(ncid, v1id, "_FillValue",
				  NC_FLOAT, 1, &nan));

    nc_close(ncid);
    if (v_new) {
      free(v_new);
      v_new = NULL;
    }
    if (v_old) {
      free(v_old);
      v_old = NULL;
    }
    if (region) {
      free(region);
      region = NULL;
    }
    if (phidp) {
      free(phidp);
      phidp = NULL;
    }
  }

  exit(0);

}
